/* ARCADEmini assembler (unix)
 * version 1.0 date 2003-09-03
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#ifndef INC_node
#define INC_node

#include <stdio.h>

#include "id_list.h"

//type of the node
typedef enum t_node_type
{
  node_generic,
  node_con,
  node_con_ident,
  node_con_unary,
  node_con_binary,
  node_con_regdif,
  node_con_bra,
  node_con_sec,
  node_con_ofs,
  node_reg_con,
  node_reg_ident,
  node_reg_binary,
  node_reg_bra,
  node_addr_ident,
  node_sec_addr,
  node_ofs_addr,
  node_reg,
  node_next_reg,
  node_str,
  node_cmd,
  node_lab,
  node_list
} et_node_type;

//data of a generic node
typedef struct t_node_generic
{
} st_node_generic;

//data of a constant node
typedef struct t_node_con
{
  unsigned long val; //value of the constant
} st_node_con;

//data of a constant identifier node
typedef struct t_node_con_ident
{
  st_id_list_entry * p_id; //pointer to the entry in the identifier table
  int val_avail; //flag if value is already available
  unsigned long val; //value of the constant
} st_node_con_ident;

//data of an unaray constant operator node
typedef struct t_node_con_unary
{
  char op; //the operator: '~', '-'
  struct t_node * p_param; //pointer to the parameter node
  int val_avail; //flag if value is already available
  unsigned long val; //value of the constant
} st_node_con_unary;

//data of a binary constant operator node
typedef struct t_node_con_binary
{
  char op; //the operator: '*', '/', '%', '+', '-', '<', '>', '&', '^', '|'
  struct t_node * p_param1, * p_param2; //pointer to the parameter nodes
  int val_avail; //flag if value is already available
  unsigned long val; //value of the constant
} st_node_con_binary;

//data of a register-difference constant operator node
typedef struct t_node_con_regdif
{
  struct t_node * p_param1, * p_param2; //pointer to the parameter nodes
  int val_avail; //flag if value is already available
  unsigned long val; //value of the constant
} st_node_con_regdif;

//data of a constant brackets node
typedef struct t_node_con_bra
{
  struct t_node * p_param; //pointer to the node within the brackets
  int val_avail; //flag if value is already available
  unsigned long val; //value of the constant
} st_node_con_bra;

//data of a constant sector node
typedef struct t_node_con_sec
{
  struct t_node * p_addr; //pointer to the address node
  int val_avail; //flag if value is already available
  unsigned long val; //value of the constant
} st_node_con_sec;

//data of a constant offset node
typedef struct t_node_con_ofs
{
  struct t_node * p_addr; //pointer to the address node
  int val_avail; //flag if value is already available
  unsigned long val; //value of the constant
} st_node_con_ofs;

//data of a register constant node
typedef struct t_node_reg_con
{
  struct t_node * p_con; //pointer to the constant node specifying the register number
  int reg_no_avail; //flag if register number is already available
  unsigned char reg_no; //number of register
} st_node_reg_con;

//data of a register identifier node
typedef struct t_node_reg_ident
{
  st_id_list_entry * p_id; //pointer to the entry in the identifier table
  int reg_no_avail; //flag if register number is already available
  unsigned char reg_no; //number of register
} st_node_reg_ident;

//data of a binary register operator node
typedef struct t_node_reg_binary
{
  char op; //the operator: '+', '-'
  struct t_node * p_reg, * p_con; //pointer to the parameter nodes
  int reg_no_avail; //flag if register number is already available
  unsigned char reg_no; //number of register
} st_node_reg_binary;

//data of a register brackets node
typedef struct t_node_reg_bra
{
  struct t_node * p_param; //pointer to the node within the brackets
  int reg_no_avail; //flag if register number is already available
  unsigned char reg_no; //number of register
} st_node_reg_bra;

//data of an address identifier node
typedef struct t_node_addr_ident
{
  st_id_list_entry * p_id; //pointer to the entry in the identifier table
  int addr_avail; //flag if address is already available
  unsigned long addr_sec; //sector part of address
  unsigned short addr_ofs; //offset part of address
} st_node_addr_ident;

//data of a sector address node (a kind of constant node)
typedef struct t_node_sec_addr
{
  struct t_node * p_addr; //pointer to the address node
  int val_avail; //flag if value is already available
  unsigned long val; //value of the constant
} st_node_sec_addr;

//data of an offset address node (a kind of constant node)
typedef struct t_node_ofs_addr
{
  struct t_node * p_addr; //pointer to the address node
  unsigned long addr_sec; //sector part of address where the address was used as offset
  int val_avail; //flag if value is already available
  unsigned long val; //value of the constant
} st_node_ofs_addr;

//data of a register node ("REGISTER")
typedef struct t_node_reg
{
  struct t_node * p_con; //pointer to the constant node specifying the register number
  int reg_no_avail; //flag if register number is already available
  unsigned char reg_no; //number of register
} st_node_reg;

//data of a next register node ("NEXT_REG")
typedef struct t_node_next_reg
{
  struct t_node * p_prev_reg; //pointer to the register node specifying the number of the previous register
  int reg_no_avail; //flag if register number is already available
  unsigned char reg_no; //number of register
} st_node_next_reg;

//parameter of a command
typedef struct t_node_cmd_param
{
  unsigned char bits; //number of bits this parameter uses
  struct t_node * p_node; //pointer to node specifying value of param
  unsigned long val; //value of parameter
                     //(if p_node is NULL)
} st_node_cmd_param;

//data of a string node
typedef struct t_node_str
{
  unsigned char len; //length of the string
  unsigned char * data; //pointer to buffer for string data
                        //(storage for these parameters is allocated
                        // behind the node structure in same memory block)
} st_node_str;

//data of a command node
typedef struct t_node_cmd
{
  unsigned char param_cnt; //number of parameters of command
                           //(first byte is also regarded as a parameter)
  struct t_node_cmd_param * params; //pointer to parameters
                                    //(storage for these parameters is allocated
                                    // behind the node structure in same memory block)
  unsigned char len; //length of command in bytes
  unsigned char * data; //pointer to buffer for command data
                        //(storage for these parameters is allocated
                        // behind the parameters in same memory block)
  int data_ok; //flag if data was already filled in
  unsigned long addr_sec; //sector part of address of command
  unsigned short addr_ofs; //offset part of address of command
} st_node_cmd;

//data of a label node
typedef struct t_node_lab
{
  struct t_id_list_entry * p_label; //pointer to the label in the id_list
  struct t_node * p_real_node; //pointer to the real node (e.g. the command node)
} st_node_lab;

//data of a list node
typedef struct t_node_list
{
  struct t_node * p_entry; //pointer to entry in list
  struct t_node * p_list; //pointer to list (part of list after this entry)
} st_node_list;

//data section of a node
typedef union t_node_data
{
  struct t_node_generic generic;
  struct t_node_con con;
  struct t_node_con_ident con_ident;
  struct t_node_con_unary con_unary;
  struct t_node_con_binary con_binary;
  struct t_node_con_regdif con_regdif;
  struct t_node_con_bra con_bra;
  struct t_node_con_sec con_sec;
  struct t_node_con_ofs con_ofs;
  struct t_node_reg_con reg_con;
  struct t_node_reg_ident reg_ident;
  struct t_node_reg_binary reg_binary;
  struct t_node_reg_bra reg_bra;
  struct t_node_addr_ident addr_ident;
  struct t_node_sec_addr sec_addr;
  struct t_node_ofs_addr ofs_addr;
  struct t_node_reg reg;
  struct t_node_next_reg next_reg;
  struct t_node_str str;
  struct t_node_cmd cmd;
  struct t_node_lab lab;
  struct t_node_list list;
} ut_node_data;

//a node
typedef struct t_node
{
  enum t_node_type type; //type of the node
  union t_node_data data; //data of the node
  unsigned long create_column, create_line; //position when node was created
  //internal used members
  struct t_node * _p_prev, * _p_next; //doubly linked linear list used to keep track of allocated nodes
} st_node;

//none register node
extern st_node node_reg_none;

//create a new generic node
st_node * node_new_generic( );

//create a new constant node
st_node * node_new_con( unsigned long val );

//create a new constant identifier node
st_node * node_new_con_ident( st_id_list_entry * p_id );

//create a new unaray constant operator node
st_node * node_new_con_unary( char op, st_node * p_param );

//create a new binary constant operator node
st_node * node_new_con_binary( char op, st_node * p_param1, st_node * p_param2 );

//create a new constant brackets node
st_node * node_new_con_bra( st_node * p_param );

//create a new constant sector node
st_node * node_new_con_sec( st_node * p_addr );

//create a new constant offset node
st_node * node_new_con_ofs( st_node * p_addr );

//create a new register-difference constant operator node
st_node * node_new_con_regdif( st_node * p_param1, st_node * p_param2 );

//create a new register constant node
st_node * node_new_reg_con( st_node * p_con );

//create a new register identifier node
st_node * node_new_reg_ident( st_id_list_entry * p_id );

//create a new binary register operator node
st_node * node_new_reg_binary( char op, st_node * p_reg, st_node * p_con );

//create a new register brackets node
st_node * node_new_reg_bra( st_node * p_param );

//create a new address identifier node
st_node * node_new_addr_ident( st_id_list_entry * p_id );

//create a new sector address node
st_node * node_new_sec_addr( st_node * p_addr );

//create a new offset address node
st_node * node_new_ofs_addr( st_node * p_addr, unsigned long cur_addr_sec );

//create a new register node ("REGISTER")
st_node * node_new_reg( st_node * p_con );

//create a new next register node ("NEXT_REG")
st_node * node_new_next_reg( st_node * p_prev_reg );

//create a new string node
st_node * node_new_str( unsigned char len );

//create a new command node
st_node * node_new_cmd( unsigned char param_cnt, unsigned char len );

//create a new label node
st_node * node_new_lab( st_id_list_entry * p_label, st_node * p_real_node );

//create a new list node
st_node * node_new_list( st_node * p_entry, st_node * p_list );

//free a node
void node_free( st_node * p_node );

//free all nodes
void node_free_all( );

#endif //#ifndef INC_node

